import logging
from . import opik_usage
from typing import Dict, Any, Callable, Optional, Union
from opik.types import LLMProvider


# TODO: it was discovered that one provider may have more than one token usage formats, and
# sometimes we support one usage format but not the cost tracking for its provider.
# So, this dictionary and corresponding logic should be updated.
# We need to have one dict - USAGE_FORMAT: OPIK_USAGE_BUILDER, and probably
# another dict - PROVIDER: USAGE_FORMATS.
# See the ticket: OPIK-1407
_PROVIDER_TO_OPIK_USAGE_BUILDER: Dict[
    Union[str, LLMProvider], Callable[[Dict[str, Any]], opik_usage.OpikUsage]
] = {
    LLMProvider.OPENAI: opik_usage.OpikUsage.from_openai_completions_dict,
    LLMProvider.GOOGLE_VERTEXAI: opik_usage.OpikUsage.from_google_dict,
    LLMProvider.GOOGLE_AI: opik_usage.OpikUsage.from_google_dict,
    LLMProvider.ANTHROPIC: opik_usage.OpikUsage.from_anthropic_dict,
    "_bedrock": opik_usage.OpikUsage.from_bedrock_dict,
    "_openai_responses": opik_usage.OpikUsage.from_openai_responses_dict,
}


def build_opik_usage(
    provider: Union[str, LLMProvider],
    usage: Dict[str, Any],
) -> opik_usage.OpikUsage:
    build_function = _PROVIDER_TO_OPIK_USAGE_BUILDER[provider]

    result = build_function(usage)

    return result


def build_opik_usage_from_unknown_provider(
    usage: Dict[str, Any],
) -> opik_usage.OpikUsage:
    for build_function in _PROVIDER_TO_OPIK_USAGE_BUILDER.values():
        try:
            opik_usage_ = build_function(usage)
            return opik_usage_
        except Exception:
            pass

    return opik_usage.OpikUsage.from_unknown_usage_dict(usage)


def try_build_opik_usage_or_log_error(
    provider: Union[str, LLMProvider],
    usage: Dict[str, Any],
    logger: logging.Logger,
    error_message: str,
) -> Optional[opik_usage.OpikUsage]:
    try:
        return build_opik_usage(provider=provider, usage=usage)
    except Exception:
        logger.error(error_message, exc_info=True)
        return None



from typing import Any, Dict, List, Optional

from opik.rest_api import client as rest_client
from opik.rest_api import core as rest_api_core
from opik.rest_api.types import prompt_version_detail, PromptVersionDetailType

from . import prompt as opik_prompt
from .prompt import PromptType


class PromptClient:
    def __init__(self, client: rest_client.OpikApi):
        self._rest_client = client

    def create_prompt(
        self,
        name: str,
        prompt: str,
        metadata: Optional[Dict[str, Any]],
        type: PromptType = PromptType.MUSTACHE,
    ) -> opik_prompt.Prompt:
        """
        Creates the prompt detail for the given prompt name and template.

        Parameters:
        - name: The name of the prompt.
        - prompt: The template content for the prompt.

        Returns:
        - A Prompt object for the provided prompt name and template.
        """
        prompt_version = self._get_latest_version(name)

        if (
            prompt_version is None
            or prompt_version.template != prompt
            or prompt_version.metadata != metadata
            or prompt_version.type != type.value
        ):
            prompt_version = self._create_new_version(
                name=name, prompt=prompt, type=type, metadata=metadata
            )

        prompt_obj = opik_prompt.Prompt.from_fern_prompt_version(
            name=name, prompt_version=prompt_version
        )

        return prompt_obj

    def _create_new_version(
        self,
        name: str,
        prompt: str,
        type: PromptVersionDetailType,
        metadata: Optional[Dict[str, Any]],
    ) -> prompt_version_detail.PromptVersionDetail:
        new_prompt_version_detail_data = prompt_version_detail.PromptVersionDetail(
            template=prompt,
            metadata=metadata,
            type=type,
        )
        new_prompt_version_detail: prompt_version_detail.PromptVersionDetail = (
            self._rest_client.prompts.create_prompt_version(
                name=name,
                version=new_prompt_version_detail_data,
            )
        )
        return new_prompt_version_detail

    def _get_latest_version(
        self, name: str
    ) -> Optional[prompt_version_detail.PromptVersionDetail]:
        try:
            prompt_latest_version = self._rest_client.prompts.retrieve_prompt_version(
                name=name
            )
            return prompt_latest_version
        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e
            return None

    def get_prompt(
        self,
        name: str,
        commit: Optional[str] = None,
    ) -> Optional[opik_prompt.Prompt]:
        """
        Retrieve the prompt detail for a given prompt name and commit version.

        Parameters:
            name: The name of the prompt.
            commit: An optional commit version of the prompt. If not provided, the latest version is retrieved.

        Returns:
            Prompt: The details of the specified prompt.
        """
        try:
            prompt_version = self._rest_client.prompts.retrieve_prompt_version(
                name=name,
                commit=commit,
            )
            prompt_obj = opik_prompt.Prompt.from_fern_prompt_version(
                name=name,
                prompt_version=prompt_version,
            )

            return prompt_obj

        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e

        return None

    # TODO: Need to add support for prompt name in the BE so we don't
    # need to retrieve the prompt id
    def get_all_prompts(self, name: str) -> List[opik_prompt.Prompt]:
        """
        Retrieve all the prompt details for a given prompt name.

        Parameters:
            name: The name of the prompt.

        Returns:
            List[Prompt]: A list of prompts for the given name.
        """
        try:
            prompts_matching_name_string = self._rest_client.prompts.get_prompts(
                name=name
            ).content
            if (
                prompts_matching_name_string is None
                or len(prompts_matching_name_string) == 0
            ):
                raise ValueError("No prompts found for name: " + name)

            filtered_prompt_list = [
                x.id for x in prompts_matching_name_string if name == x.name
            ]
            if len(filtered_prompt_list) == 0:
                raise ValueError("No prompts found for name: " + name)

            prompt_id = filtered_prompt_list[0]

            page = 1
            size = 100

            prompts: List[opik_prompt.Prompt] = []
            while True:
                prompt_versions = self._rest_client.prompts.get_prompt_versions(
                    id=prompt_id, page=page, size=size
                ).content
                prompts.extend(
                    [
                        opik_prompt.Prompt.from_fern_prompt_version(name, version)
                        for version in prompt_versions
                    ]
                )
                if len(prompt_versions) < size:
                    break
                page += 1

            return prompts

        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e

        return []



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .optimization_update_status import OptimizationUpdateStatus
from .optimization_write_status import OptimizationWriteStatus

__all__ = ["OptimizationUpdateStatus", "OptimizationWriteStatus"]



import openai


def is_llm_provider_rate_limit_error(exception: Exception) -> bool:
    import litellm.exceptions

    rate_limit_error_known_types = (
        openai.RateLimitError,
        litellm.exceptions.RateLimitError,
    )

    is_rate_limit_error = isinstance(exception, rate_limit_error_known_types) or (
        hasattr(exception, "status_code") and exception.status_code == 429
    )

    return is_rate_limit_error



import logging
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union, Literal

from opik import _logging as opik_logging
from opik import llm_usage, logging_messages

if TYPE_CHECKING:
    from langchain_core.tracers.schemas import Run


LOGGER = logging.getLogger(__name__)


def get_llm_usage_info(
    run_dict: Optional[Dict[str, Any]] = None,
) -> llm_usage.LLMUsageInfo:
    if run_dict is None:
        return llm_usage.LLMUsageInfo()

    opik_usage = _try_get_token_usage(run_dict)
    provider, model = _get_provider_and_model(run_dict)

    return llm_usage.LLMUsageInfo(provider=provider, model=model, usage=opik_usage)


def _try_get_token_usage(run_dict: Dict[str, Any]) -> Optional[llm_usage.OpikUsage]:
    """
    Attempts to extract and return the token usage from the given run dictionary.

    Depending on the execution type (invoke, streaming mode, async, etc.), or even the model name itself,
    token usage info might be in different places, different formats, or completely missing.
    """
    try:
        if run_dict["outputs"]["llm_output"] is not None:
            token_usage_dict = run_dict["outputs"]["llm_output"]["token_usage"]
            if not isinstance(token_usage_dict, dict):
                return None
            return llm_usage.OpikUsage.from_openai_completions_dict(token_usage_dict)

        # streaming mode handling
        # token usage data MAY be available at the end of streaming
        # in async mode may not provide token usage info
        elif token_usage_dict := run_dict["outputs"]["generations"][-1][-1]["message"][
            "kwargs"
        ].get("usage_metadata"):
            # TODO: provide better support for langchain usage. We probably need to convert it
            # to the full openai usage dict (with details) or find a way to access raw data
            openai_formatted_dict = {
                "completion_tokens": token_usage_dict["output_tokens"],
                "prompt_tokens": token_usage_dict["input_tokens"],
                "total_tokens": token_usage_dict["total_tokens"],
            }
            opik_usage = llm_usage.OpikUsage.from_openai_completions_dict(
                openai_formatted_dict
            )
            return opik_usage
        else:
            opik_logging.log_once_at_level(
                logging_level=logging.WARNING,
                message=logging_messages.WARNING_TOKEN_USAGE_DATA_IS_NOT_AVAILABLE,
                logger=LOGGER,
            )
            return None

    except Exception:
        LOGGER.warning(
            logging_messages.FAILED_TO_EXTRACT_TOKEN_USAGE_FROM_PRESUMABLY_LANGCHAIN_OPENAI_LLM_RUN,
            exc_info=True,
        )
        return None


def is_openai_run(run: "Run") -> bool:
    try:
        if run.serialized is None:
            return False

        serialized_kwargs = run.serialized.get("kwargs", {})
        has_openai_key = "openai_api_key" in serialized_kwargs

        return has_openai_key

    except Exception:
        LOGGER.debug(
            "Failed to check if Run instance is from OpenAI LLM, returning False.",
            exc_info=True,
        )
        return False


def _get_provider_and_model(
    run_dict: Dict[str, Any],
) -> Tuple[Optional[Union[Literal["openai"], str]], Optional[str]]:
    """
    Fetches the provider and model information from a given run dictionary.

    By default, the provider is assumed to be OpenAI (will be available in extra/metadata field).
    If LLM output is available, the model version is included.
    If the Client is available, the BaseURL is also checked.
    """
    provider = None
    model = None

    # here will be always provider and model
    # but model name may not be full (model+version)
    if metadata := run_dict["extra"].get("metadata"):
        provider = metadata.get("ls_provider")
        model = metadata.get("ls_model_name")

    # try to detect model+version more precise way if possible
    # .invoke() mode
    if llm_output := run_dict["outputs"].get("llm_output"):
        model = llm_output.get("model_name", model)
    # streaming mode
    elif generation_info := run_dict["outputs"]["generations"][-1][-1][
        "generation_info"
    ]:
        model = generation_info.get("model_name", model)

    # provider: check base url
    if base_url := run_dict["extra"].get("invocation_params", {}).get("base_url"):
        if base_url.host != "api.openai.com":
            provider = base_url.host

    return provider, model



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.not_found_error import NotFoundError
from ..types.json_node_write import JsonNodeWrite
from ..types.optimization_page_public import OptimizationPagePublic
from ..types.optimization_public import OptimizationPublic
from .types.optimization_update_status import OptimizationUpdateStatus
from .types.optimization_write_status import OptimizationWriteStatus

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawOptimizationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def find_optimizations(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[OptimizationPagePublic]:
        """
        Find optimizations

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OptimizationPagePublic]
            Optimizations resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/optimizations",
            method="GET",
            params={
                "page": page,
                "size": size,
                "dataset_id": dataset_id,
                "name": name,
                "dataset_deleted": dataset_deleted,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OptimizationPagePublic,
                    parse_obj_as(
                        type_=OptimizationPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_optimization(
        self,
        *,
        dataset_name: str,
        objective_name: str,
        status: OptimizationWriteStatus,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create optimization

        Parameters
        ----------
        dataset_name : str

        objective_name : str

        status : OptimizationWriteStatus

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/optimizations",
            method="POST",
            json={
                "id": id,
                "name": name,
                "dataset_name": dataset_name,
                "objective_name": objective_name,
                "status": status,
                "metadata": metadata,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_optimizations_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete optimizations by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/optimizations/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_optimization_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[OptimizationPublic]:
        """
        Get optimization by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[OptimizationPublic]
            Optimization resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/optimizations/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OptimizationPublic,
                    parse_obj_as(
                        type_=OptimizationPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_optimizations_by_id(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        status: typing.Optional[OptimizationUpdateStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update optimization by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        status : typing.Optional[OptimizationUpdateStatus]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/optimizations/{jsonable_encoder(id)}",
            method="PUT",
            json={
                "name": name,
                "status": status,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawOptimizationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def find_optimizations(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[OptimizationPagePublic]:
        """
        Find optimizations

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OptimizationPagePublic]
            Optimizations resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/optimizations",
            method="GET",
            params={
                "page": page,
                "size": size,
                "dataset_id": dataset_id,
                "name": name,
                "dataset_deleted": dataset_deleted,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OptimizationPagePublic,
                    parse_obj_as(
                        type_=OptimizationPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_optimization(
        self,
        *,
        dataset_name: str,
        objective_name: str,
        status: OptimizationWriteStatus,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create optimization

        Parameters
        ----------
        dataset_name : str

        objective_name : str

        status : OptimizationWriteStatus

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/optimizations",
            method="POST",
            json={
                "id": id,
                "name": name,
                "dataset_name": dataset_name,
                "objective_name": objective_name,
                "status": status,
                "metadata": metadata,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_optimizations_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete optimizations by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/optimizations/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_optimization_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[OptimizationPublic]:
        """
        Get optimization by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[OptimizationPublic]
            Optimization resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/optimizations/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    OptimizationPublic,
                    parse_obj_as(
                        type_=OptimizationPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_optimizations_by_id(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        status: typing.Optional[OptimizationUpdateStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update optimization by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        status : typing.Optional[OptimizationUpdateStatus]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/optimizations/{jsonable_encoder(id)}",
            method="PUT",
            json={
                "name": name,
                "status": status,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .categorical_feedback_detail_create import CategoricalFeedbackDetailCreate
from .numerical_feedback_detail_create import NumericalFeedbackDetailCreate


class Base(UniversalBaseModel):
    id: typing.Optional[str] = None
    name: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class FeedbackCreate_Numerical(Base):
    type: typing.Literal["numerical"] = "numerical"
    details: typing.Optional[NumericalFeedbackDetailCreate] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class FeedbackCreate_Categorical(Base):
    type: typing.Literal["categorical"] = "categorical"
    details: typing.Optional[CategoricalFeedbackDetailCreate] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


FeedbackCreate = typing.Union[FeedbackCreate_Numerical, FeedbackCreate_Categorical]



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.not_found_error import NotFoundError


class RawRedirectClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def datasets_redirect(
        self,
        *,
        dataset_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create dataset redirect url

        Parameters
        ----------
        dataset_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/session/redirect/datasets",
            method="GET",
            params={
                "dataset_id": dataset_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def experiments_redirect(
        self,
        *,
        dataset_id: str,
        experiment_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create experiment redirect url

        Parameters
        ----------
        dataset_id : str

        experiment_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/session/redirect/experiments",
            method="GET",
            params={
                "dataset_id": dataset_id,
                "experiment_id": experiment_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def projects_redirect(
        self,
        *,
        trace_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create project redirect url

        Parameters
        ----------
        trace_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/session/redirect/projects",
            method="GET",
            params={
                "trace_id": trace_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawRedirectClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def datasets_redirect(
        self,
        *,
        dataset_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create dataset redirect url

        Parameters
        ----------
        dataset_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/session/redirect/datasets",
            method="GET",
            params={
                "dataset_id": dataset_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def experiments_redirect(
        self,
        *,
        dataset_id: str,
        experiment_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create experiment redirect url

        Parameters
        ----------
        dataset_id : str

        experiment_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/session/redirect/experiments",
            method="GET",
            params={
                "dataset_id": dataset_id,
                "experiment_id": experiment_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def projects_redirect(
        self,
        *,
        trace_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create project redirect url

        Parameters
        ----------
        trace_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/session/redirect/projects",
            method="GET",
            params={
                "trace_id": trace_id,
                "workspace_name": workspace_name,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import logging
from opik import exceptions, logging_messages
from opik.evaluation.metrics import score_result
from opik.evaluation.metrics.llm_judges import parsing_helpers

LOGGER = logging.getLogger(__name__)


def parse_model_output(content: str, name: str) -> score_result.ScoreResult:
    try:
        dict_content = parsing_helpers.extract_json_content_or_raise(content)
        score: float = dict_content["answer_relevance_score"]

        if not (0.0 <= score <= 1.0):
            raise exceptions.MetricComputationError(
                f"Answer relevance score must be between 0.0 and 1.0, got {score}"
            )

        return score_result.ScoreResult(
            name=name, value=score, reason=dict_content["reason"]
        )
    except Exception as e:
        LOGGER.error(f"Failed to parse model output: {e}", exc_info=True)
        raise exceptions.MetricComputationError(
            logging_messages.ANSWER_RELEVANCE_SCORE_CALC_FAILED
        )



from typing import List, Optional

from opik.api_objects import experiment, opik_client
from opik.types import FeedbackScoreDict
from . import test_case, test_result
from .metrics import arguments_helpers
from .types import ScoringKeyMappingType


def get_experiment_with_unique_name(
    client: opik_client.Opik, experiment_name: str
) -> experiment.Experiment:
    experiments = client.get_experiments_by_name(name=experiment_name)

    if len(experiments) == 0:
        raise ValueError(f"Experiment with name {experiment_name} not found")
    elif len(experiments) > 1:
        raise ValueError(
            f"Found multiple experiments with name {experiment_name}. Try to use `experiment_id` instead"
        )

    experiment_ = experiments[0]
    return experiment_


def get_trace_project_name(client: opik_client.Opik, trace_id: str) -> str:
    # We first need to get the project_id for the trace
    traces = client.get_trace_content(id=trace_id)
    project_id = traces.project_id

    # Then we can get the project name
    project_metadata = client.get_project(id=project_id)
    return project_metadata.name


def get_experiment_test_cases(
    client: opik_client.Opik,
    experiment_id: str,
    dataset_id: str,
    scoring_key_mapping: Optional[ScoringKeyMappingType],
) -> List[test_case.TestCase]:
    test_cases = []
    page = 1

    while True:
        experiment_items_page = (
            client._rest_client.datasets.find_dataset_items_with_experiment_items(
                id=dataset_id, experiment_ids=f'["{experiment_id}"]', page=page
            )
        )
        if len(experiment_items_page.content) == 0:
            break

        for item in experiment_items_page.content:
            experiment_item = item.experiment_items[0]

            test_cases += [
                test_case.TestCase(
                    trace_id=experiment_item.trace_id,
                    dataset_item_id=experiment_item.dataset_item_id,
                    task_output=experiment_item.output,
                    scoring_inputs=arguments_helpers.create_scoring_inputs(
                        dataset_item=experiment_item.input,
                        task_output=experiment_item.output,
                        scoring_key_mapping=scoring_key_mapping,
                    ),
                )
            ]

        page += 1

    return test_cases


def log_test_result_scores(
    client: opik_client.Opik,
    test_result: test_result.TestResult,
    project_name: Optional[str],
) -> None:
    all_trace_scores: List[FeedbackScoreDict] = []

    for score_result in test_result.score_results:
        if score_result.scoring_failed:
            continue

        trace_score = FeedbackScoreDict(
            id=test_result.test_case.trace_id,
            name=score_result.name,
            value=score_result.value,
            reason=score_result.reason,
        )
        all_trace_scores.append(trace_score)

    client.log_traces_feedback_scores(
        scores=all_trace_scores, project_name=project_name
    )



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import (
    AssistantMessage,
    AssistantMessageRole,
    Attachment,
    AttachmentPage,
    AuthDetailsHolder,
    AutomationRuleEvaluator,
    AutomationRuleEvaluatorLlmAsJudge,
    AutomationRuleEvaluatorLlmAsJudgePublic,
    AutomationRuleEvaluatorLlmAsJudgeWrite,
    AutomationRuleEvaluatorObjectPublic,
    AutomationRuleEvaluatorObjectPublic_LlmAsJudge,
    AutomationRuleEvaluatorObjectPublic_UserDefinedMetricPython,
    AutomationRuleEvaluatorPagePublic,
    AutomationRuleEvaluatorPublic,
    AutomationRuleEvaluatorPublic_LlmAsJudge,
    AutomationRuleEvaluatorPublic_UserDefinedMetricPython,
    AutomationRuleEvaluatorUpdate,
    AutomationRuleEvaluatorUpdateLlmAsJudge,
    AutomationRuleEvaluatorUpdateUserDefinedMetricPython,
    AutomationRuleEvaluatorUpdate_LlmAsJudge,
    AutomationRuleEvaluatorUpdate_UserDefinedMetricPython,
    AutomationRuleEvaluatorUserDefinedMetricPython,
    AutomationRuleEvaluatorUserDefinedMetricPythonPublic,
    AutomationRuleEvaluatorUserDefinedMetricPythonWrite,
    AutomationRuleEvaluatorWrite,
    AutomationRuleEvaluatorWrite_LlmAsJudge,
    AutomationRuleEvaluatorWrite_UserDefinedMetricPython,
    AutomationRuleEvaluator_LlmAsJudge,
    AutomationRuleEvaluator_UserDefinedMetricPython,
    AvgValueStatPublic,
    BatchDelete,
    BiInformation,
    BiInformationResponse,
    CategoricalFeedbackDefinition,
    CategoricalFeedbackDefinitionCreate,
    CategoricalFeedbackDefinitionPublic,
    CategoricalFeedbackDefinitionUpdate,
    CategoricalFeedbackDetail,
    CategoricalFeedbackDetailCreate,
    CategoricalFeedbackDetailPublic,
    CategoricalFeedbackDetailUpdate,
    ChatCompletionChoice,
    ChatCompletionResponse,
    Check,
    CheckName,
    CheckPublic,
    CheckPublicName,
    CheckPublicResult,
    CheckResult,
    ChunkedOutputJsonNode,
    ChunkedOutputJsonNodePublic,
    ChunkedOutputJsonNodePublicType,
    ChunkedOutputJsonNodeType,
    Column,
    ColumnCompare,
    ColumnCompareTypesItem,
    ColumnPublic,
    ColumnPublicTypesItem,
    ColumnTypesItem,
    Comment,
    CommentCompare,
    CommentPublic,
    CompleteMultipartUploadRequest,
    CompleteMultipartUploadRequestEntityType,
    CompletionTokensDetails,
    CountValueStatPublic,
    DataPointNumberPublic,
    Dataset,
    DatasetItem,
    DatasetItemBatch,
    DatasetItemCompare,
    DatasetItemCompareSource,
    DatasetItemPageCompare,
    DatasetItemPagePublic,
    DatasetItemPublic,
    DatasetItemPublicSource,
    DatasetItemSource,
    DatasetItemWrite,
    DatasetItemWriteSource,
    DatasetPagePublic,
    DatasetPublic,
    DatasetPublicVisibility,
    DatasetVisibility,
    DeleteAttachmentsRequest,
    DeleteAttachmentsRequestEntityType,
    DeleteFeedbackScore,
    DeleteIdsHolder,
    Delta,
    DeltaRole,
    ErrorInfo,
    ErrorInfoPublic,
    ErrorInfoWrite,
    ErrorMessage,
    ErrorMessageDetail,
    ErrorMessageDetailed,
    ErrorMessagePublic,
    Experiment,
    ExperimentItem,
    ExperimentItemCompare,
    ExperimentItemPublic,
    ExperimentPagePublic,
    ExperimentPublic,
    ExperimentPublicType,
    ExperimentType,
    ExportTraceServiceRequest,
    Feedback,
    FeedbackCreate,
    FeedbackCreate_Categorical,
    FeedbackCreate_Numerical,
    FeedbackDefinitionPagePublic,
    FeedbackObjectPublic,
    FeedbackObjectPublic_Categorical,
    FeedbackObjectPublic_Numerical,
    FeedbackPublic,
    FeedbackPublic_Categorical,
    FeedbackPublic_Numerical,
    FeedbackScore,
    FeedbackScoreAverage,
    FeedbackScoreAverageDetailed,
    FeedbackScoreAveragePublic,
    FeedbackScoreBatch,
    FeedbackScoreBatchItem,
    FeedbackScoreBatchItemSource,
    FeedbackScoreCompare,
    FeedbackScoreCompareSource,
    FeedbackScoreNames,
    FeedbackScorePublic,
    FeedbackScorePublicSource,
    FeedbackScoreSource,
    FeedbackUpdate,
    FeedbackUpdate_Categorical,
    FeedbackUpdate_Numerical,
    Feedback_Categorical,
    Feedback_Numerical,
    Function,
    FunctionCall,
    Guardrail,
    GuardrailBatch,
    GuardrailName,
    GuardrailResult,
    GuardrailWrite,
    GuardrailWriteName,
    GuardrailWriteResult,
    GuardrailsValidation,
    GuardrailsValidationPublic,
    JsonNode,
    JsonNodeCompare,
    JsonNodeDetail,
    JsonNodePublic,
    JsonNodeWrite,
    JsonObjectSchema,
    JsonSchema,
    JsonSchemaElement,
    LlmAsJudgeCode,
    LlmAsJudgeCodePublic,
    LlmAsJudgeCodeWrite,
    LlmAsJudgeMessage,
    LlmAsJudgeMessagePublic,
    LlmAsJudgeMessagePublicRole,
    LlmAsJudgeMessageRole,
    LlmAsJudgeMessageWrite,
    LlmAsJudgeMessageWriteRole,
    LlmAsJudgeModelParameters,
    LlmAsJudgeModelParametersPublic,
    LlmAsJudgeModelParametersWrite,
    LlmAsJudgeOutputSchema,
    LlmAsJudgeOutputSchemaPublic,
    LlmAsJudgeOutputSchemaPublicType,
    LlmAsJudgeOutputSchemaType,
    LlmAsJudgeOutputSchemaWrite,
    LlmAsJudgeOutputSchemaWriteType,
    LogItem,
    LogItemLevel,
    LogPage,
    Message,
    MultipartUploadPart,
    NumericalFeedbackDefinition,
    NumericalFeedbackDefinitionCreate,
    NumericalFeedbackDefinitionPublic,
    NumericalFeedbackDefinitionUpdate,
    NumericalFeedbackDetail,
    NumericalFeedbackDetailCreate,
    NumericalFeedbackDetailPublic,
    NumericalFeedbackDetailUpdate,
    Optimization,
    OptimizationPagePublic,
    OptimizationPublic,
    OptimizationPublicStatus,
    OptimizationStatus,
    PageColumns,
    PercentageValueStatPublic,
    PercentageValues,
    PercentageValuesDetailed,
    PercentageValuesPublic,
    Project,
    ProjectDetailed,
    ProjectDetailedVisibility,
    ProjectMetricResponsePublic,
    ProjectMetricResponsePublicInterval,
    ProjectMetricResponsePublicMetricType,
    ProjectPagePublic,
    ProjectPublic,
    ProjectPublicVisibility,
    ProjectStatItemObjectPublic,
    ProjectStatItemObjectPublic_Avg,
    ProjectStatItemObjectPublic_Count,
    ProjectStatItemObjectPublic_Percentage,
    ProjectStatsPublic,
    ProjectStatsSummary,
    ProjectStatsSummaryItem,
    ProjectVisibility,
    Prompt,
    PromptDetail,
    PromptPagePublic,
    PromptPublic,
    PromptType,
    PromptVersion,
    PromptVersionDetail,
    PromptVersionDetailType,
    PromptVersionLink,
    PromptVersionLinkPublic,
    PromptVersionLinkWrite,
    PromptVersionPagePublic,
    PromptVersionPublic,
    PromptVersionPublicType,
    PromptVersionType,
    ProviderApiKey,
    ProviderApiKeyPagePublic,
    ProviderApiKeyProvider,
    ProviderApiKeyPublic,
    ProviderApiKeyPublicProvider,
    ResponseFormat,
    ResponseFormatType,
    ResultsNumberPublic,
    ScoreName,
    ServiceTogglesConfig,
    Span,
    SpanBatch,
    SpanFilterPublic,
    SpanFilterPublicOperator,
    SpanPagePublic,
    SpanPublic,
    SpanPublicType,
    SpanType,
    SpanWrite,
    SpanWriteType,
    SpansCountResponse,
    StartMultipartUploadResponse,
    StreamOptions,
    Tool,
    ToolCall,
    Trace,
    TraceBatch,
    TraceCountResponse,
    TraceFilterPublic,
    TraceFilterPublicOperator,
    TracePagePublic,
    TracePublic,
    TraceThread,
    TraceThreadPage,
    TraceWrite,
    Usage,
    UserDefinedMetricPythonCode,
    UserDefinedMetricPythonCodePublic,
    UserDefinedMetricPythonCodeWrite,
    WorkspaceNameHolder,
    WorkspaceSpansCount,
    WorkspaceTraceCount,
)
from .errors import (
    BadRequestError,
    ConflictError,
    ForbiddenError,
    NotFoundError,
    NotImplementedError,
    UnauthorizedError,
    UnprocessableEntityError,
)
from . import (
    attachments,
    automation_rule_evaluators,
    chat_completions,
    check,
    datasets,
    experiments,
    feedback_definitions,
    guardrails,
    llm_provider_key,
    open_telemetry_ingestion,
    optimizations,
    projects,
    prompts,
    redirect,
    service_toggles,
    spans,
    system_usage,
    traces,
)
from .attachments import (
    AttachmentListRequestEntityType,
    DownloadAttachmentRequestEntityType,
    StartMultipartUploadRequestEntityType,
    UploadAttachmentRequestEntityType,
)
from .client import AsyncOpikApi, OpikApi
from .datasets import DatasetUpdateVisibility, DatasetWriteVisibility
from .environment import OpikApiEnvironment
from .experiments import ExperimentWriteType
from .feedback_definitions import FindFeedbackDefinitionsRequestType
from .llm_provider_key import ProviderApiKeyWriteProvider
from .optimizations import OptimizationUpdateStatus, OptimizationWriteStatus
from .projects import (
    ProjectMetricRequestPublicInterval,
    ProjectMetricRequestPublicMetricType,
    ProjectUpdateVisibility,
    ProjectWriteVisibility,
)
from .prompts import PromptWriteType
from .spans import (
    FindFeedbackScoreNames1RequestType,
    GetSpanStatsRequestType,
    GetSpansByProjectRequestType,
    SpanSearchStreamRequestPublicType,
)

__all__ = [
    "AssistantMessage",
    "AssistantMessageRole",
    "AsyncOpikApi",
    "Attachment",
    "AttachmentListRequestEntityType",
    "AttachmentPage",
    "AuthDetailsHolder",
    "AutomationRuleEvaluator",
    "AutomationRuleEvaluatorLlmAsJudge",
    "AutomationRuleEvaluatorLlmAsJudgePublic",
    "AutomationRuleEvaluatorLlmAsJudgeWrite",
    "AutomationRuleEvaluatorObjectPublic",
    "AutomationRuleEvaluatorObjectPublic_LlmAsJudge",
    "AutomationRuleEvaluatorObjectPublic_UserDefinedMetricPython",
    "AutomationRuleEvaluatorPagePublic",
    "AutomationRuleEvaluatorPublic",
    "AutomationRuleEvaluatorPublic_LlmAsJudge",
    "AutomationRuleEvaluatorPublic_UserDefinedMetricPython",
    "AutomationRuleEvaluatorUpdate",
    "AutomationRuleEvaluatorUpdateLlmAsJudge",
    "AutomationRuleEvaluatorUpdateUserDefinedMetricPython",
    "AutomationRuleEvaluatorUpdate_LlmAsJudge",
    "AutomationRuleEvaluatorUpdate_UserDefinedMetricPython",
    "AutomationRuleEvaluatorUserDefinedMetricPython",
    "AutomationRuleEvaluatorUserDefinedMetricPythonPublic",
    "AutomationRuleEvaluatorUserDefinedMetricPythonWrite",
    "AutomationRuleEvaluatorWrite",
    "AutomationRuleEvaluatorWrite_LlmAsJudge",
    "AutomationRuleEvaluatorWrite_UserDefinedMetricPython",
    "AutomationRuleEvaluator_LlmAsJudge",
    "AutomationRuleEvaluator_UserDefinedMetricPython",
    "AvgValueStatPublic",
    "BadRequestError",
    "BatchDelete",
    "BiInformation",
    "BiInformationResponse",
    "CategoricalFeedbackDefinition",
    "CategoricalFeedbackDefinitionCreate",
    "CategoricalFeedbackDefinitionPublic",
    "CategoricalFeedbackDefinitionUpdate",
    "CategoricalFeedbackDetail",
    "CategoricalFeedbackDetailCreate",
    "CategoricalFeedbackDetailPublic",
    "CategoricalFeedbackDetailUpdate",
    "ChatCompletionChoice",
    "ChatCompletionResponse",
    "Check",
    "CheckName",
    "CheckPublic",
    "CheckPublicName",
    "CheckPublicResult",
    "CheckResult",
    "ChunkedOutputJsonNode",
    "ChunkedOutputJsonNodePublic",
    "ChunkedOutputJsonNodePublicType",
    "ChunkedOutputJsonNodeType",
    "Column",
    "ColumnCompare",
    "ColumnCompareTypesItem",
    "ColumnPublic",
    "ColumnPublicTypesItem",
    "ColumnTypesItem",
    "Comment",
    "CommentCompare",
    "CommentPublic",
    "CompleteMultipartUploadRequest",
    "CompleteMultipartUploadRequestEntityType",
    "CompletionTokensDetails",
    "ConflictError",
    "CountValueStatPublic",
    "DataPointNumberPublic",
    "Dataset",
    "DatasetItem",
    "DatasetItemBatch",
    "DatasetItemCompare",
    "DatasetItemCompareSource",
    "DatasetItemPageCompare",
    "DatasetItemPagePublic",
    "DatasetItemPublic",
    "DatasetItemPublicSource",
    "DatasetItemSource",
    "DatasetItemWrite",
    "DatasetItemWriteSource",
    "DatasetPagePublic",
    "DatasetPublic",
    "DatasetPublicVisibility",
    "DatasetUpdateVisibility",
    "DatasetVisibility",
    "DatasetWriteVisibility",
    "DeleteAttachmentsRequest",
    "DeleteAttachmentsRequestEntityType",
    "DeleteFeedbackScore",
    "DeleteIdsHolder",
    "Delta",
    "DeltaRole",
    "DownloadAttachmentRequestEntityType",
    "ErrorInfo",
    "ErrorInfoPublic",
    "ErrorInfoWrite",
    "ErrorMessage",
    "ErrorMessageDetail",
    "ErrorMessageDetailed",
    "ErrorMessagePublic",
    "Experiment",
    "ExperimentItem",
    "ExperimentItemCompare",
    "ExperimentItemPublic",
    "ExperimentPagePublic",
    "ExperimentPublic",
    "ExperimentPublicType",
    "ExperimentType",
    "ExperimentWriteType",
    "ExportTraceServiceRequest",
    "Feedback",
    "FeedbackCreate",
    "FeedbackCreate_Categorical",
    "FeedbackCreate_Numerical",
    "FeedbackDefinitionPagePublic",
    "FeedbackObjectPublic",
    "FeedbackObjectPublic_Categorical",
    "FeedbackObjectPublic_Numerical",
    "FeedbackPublic",
    "FeedbackPublic_Categorical",
    "FeedbackPublic_Numerical",
    "FeedbackScore",
    "FeedbackScoreAverage",
    "FeedbackScoreAverageDetailed",
    "FeedbackScoreAveragePublic",
    "FeedbackScoreBatch",
    "FeedbackScoreBatchItem",
    "FeedbackScoreBatchItemSource",
    "FeedbackScoreCompare",
    "FeedbackScoreCompareSource",
    "FeedbackScoreNames",
    "FeedbackScorePublic",
    "FeedbackScorePublicSource",
    "FeedbackScoreSource",
    "FeedbackUpdate",
    "FeedbackUpdate_Categorical",
    "FeedbackUpdate_Numerical",
    "Feedback_Categorical",
    "Feedback_Numerical",
    "FindFeedbackDefinitionsRequestType",
    "FindFeedbackScoreNames1RequestType",
    "ForbiddenError",
    "Function",
    "FunctionCall",
    "GetSpanStatsRequestType",
    "GetSpansByProjectRequestType",
    "Guardrail",
    "GuardrailBatch",
    "GuardrailName",
    "GuardrailResult",
    "GuardrailWrite",
    "GuardrailWriteName",
    "GuardrailWriteResult",
    "GuardrailsValidation",
    "GuardrailsValidationPublic",
    "JsonNode",
    "JsonNodeCompare",
    "JsonNodeDetail",
    "JsonNodePublic",
    "JsonNodeWrite",
    "JsonObjectSchema",
    "JsonSchema",
    "JsonSchemaElement",
    "LlmAsJudgeCode",
    "LlmAsJudgeCodePublic",
    "LlmAsJudgeCodeWrite",
    "LlmAsJudgeMessage",
    "LlmAsJudgeMessagePublic",
    "LlmAsJudgeMessagePublicRole",
    "LlmAsJudgeMessageRole",
    "LlmAsJudgeMessageWrite",
    "LlmAsJudgeMessageWriteRole",
    "LlmAsJudgeModelParameters",
    "LlmAsJudgeModelParametersPublic",
    "LlmAsJudgeModelParametersWrite",
    "LlmAsJudgeOutputSchema",
    "LlmAsJudgeOutputSchemaPublic",
    "LlmAsJudgeOutputSchemaPublicType",
    "LlmAsJudgeOutputSchemaType",
    "LlmAsJudgeOutputSchemaWrite",
    "LlmAsJudgeOutputSchemaWriteType",
    "LogItem",
    "LogItemLevel",
    "LogPage",
    "Message",
    "MultipartUploadPart",
    "NotFoundError",
    "NotImplementedError",
    "NumericalFeedbackDefinition",
    "NumericalFeedbackDefinitionCreate",
    "NumericalFeedbackDefinitionPublic",
    "NumericalFeedbackDefinitionUpdate",
    "NumericalFeedbackDetail",
    "NumericalFeedbackDetailCreate",
    "NumericalFeedbackDetailPublic",
    "NumericalFeedbackDetailUpdate",
    "OpikApi",
    "OpikApiEnvironment",
    "Optimization",
    "OptimizationPagePublic",
    "OptimizationPublic",
    "OptimizationPublicStatus",
    "OptimizationStatus",
    "OptimizationUpdateStatus",
    "OptimizationWriteStatus",
    "PageColumns",
    "PercentageValueStatPublic",
    "PercentageValues",
    "PercentageValuesDetailed",
    "PercentageValuesPublic",
    "Project",
    "ProjectDetailed",
    "ProjectDetailedVisibility",
    "ProjectMetricRequestPublicInterval",
    "ProjectMetricRequestPublicMetricType",
    "ProjectMetricResponsePublic",
    "ProjectMetricResponsePublicInterval",
    "ProjectMetricResponsePublicMetricType",
    "ProjectPagePublic",
    "ProjectPublic",
    "ProjectPublicVisibility",
    "ProjectStatItemObjectPublic",
    "ProjectStatItemObjectPublic_Avg",
    "ProjectStatItemObjectPublic_Count",
    "ProjectStatItemObjectPublic_Percentage",
    "ProjectStatsPublic",
    "ProjectStatsSummary",
    "ProjectStatsSummaryItem",
    "ProjectUpdateVisibility",
    "ProjectVisibility",
    "ProjectWriteVisibility",
    "Prompt",
    "PromptDetail",
    "PromptPagePublic",
    "PromptPublic",
    "PromptType",
    "PromptVersion",
    "PromptVersionDetail",
    "PromptVersionDetailType",
    "PromptVersionLink",
    "PromptVersionLinkPublic",
    "PromptVersionLinkWrite",
    "PromptVersionPagePublic",
    "PromptVersionPublic",
    "PromptVersionPublicType",
    "PromptVersionType",
    "PromptWriteType",
    "ProviderApiKey",
    "ProviderApiKeyPagePublic",
    "ProviderApiKeyProvider",
    "ProviderApiKeyPublic",
    "ProviderApiKeyPublicProvider",
    "ProviderApiKeyWriteProvider",
    "ResponseFormat",
    "ResponseFormatType",
    "ResultsNumberPublic",
    "ScoreName",
    "ServiceTogglesConfig",
    "Span",
    "SpanBatch",
    "SpanFilterPublic",
    "SpanFilterPublicOperator",
    "SpanPagePublic",
    "SpanPublic",
    "SpanPublicType",
    "SpanSearchStreamRequestPublicType",
    "SpanType",
    "SpanWrite",
    "SpanWriteType",
    "SpansCountResponse",
    "StartMultipartUploadRequestEntityType",
    "StartMultipartUploadResponse",
    "StreamOptions",
    "Tool",
    "ToolCall",
    "Trace",
    "TraceBatch",
    "TraceCountResponse",
    "TraceFilterPublic",
    "TraceFilterPublicOperator",
    "TracePagePublic",
    "TracePublic",
    "TraceThread",
    "TraceThreadPage",
    "TraceWrite",
    "UnauthorizedError",
    "UnprocessableEntityError",
    "UploadAttachmentRequestEntityType",
    "Usage",
    "UserDefinedMetricPythonCode",
    "UserDefinedMetricPythonCodePublic",
    "UserDefinedMetricPythonCodeWrite",
    "WorkspaceNameHolder",
    "WorkspaceSpansCount",
    "WorkspaceTraceCount",
    "attachments",
    "automation_rule_evaluators",
    "chat_completions",
    "check",
    "datasets",
    "experiments",
    "feedback_definitions",
    "guardrails",
    "llm_provider_key",
    "open_telemetry_ingestion",
    "optimizations",
    "projects",
    "prompts",
    "redirect",
    "service_toggles",
    "spans",
    "system_usage",
    "traces",
]



from datetime import datetime
from typing import Optional
import uuid
import uuid6


def generate_id(timestamp: Optional[datetime] = None) -> str:
    if timestamp:
        uuid4 = str(uuid.uuid4())
        return str(uuid4_to_uuid7(timestamp, uuid4))

    return str(uuid6.uuid7())


def uuid4_to_uuid7(user_datetime: datetime, user_uuid: str) -> uuid.UUID:
    """Convert a UUID v4 into a UUID v7 following RFC draft specification."""
    # Get Unix timestamp in milliseconds
    unix_ts_ms = int(user_datetime.timestamp() * 1000)

    uuidv4 = uuid.UUID(user_uuid)
    if uuidv4.version != 4:
        raise ValueError("Input UUID must be version 4")

    # Create the 16-byte array
    uuid_bytes = bytearray(16)

    # First 48 bits (6 bytes): Unix timestamp in milliseconds
    uuid_bytes[0:6] = unix_ts_ms.to_bytes(6, byteorder="big")

    # Next byte: Version 7 in top 4 bits
    uuid_bytes[6] = 0x70 | (uuidv4.bytes[6] & 0x0F)

    # Next byte: random from UUID v4
    uuid_bytes[7] = uuidv4.bytes[7]

    # Next byte: Variant bits (0b10) in top 2 bits
    uuid_bytes[8] = 0x80 | (uuidv4.bytes[8] & 0x3F)

    # Remaining bytes: random from UUID v4
    uuid_bytes[9:16] = uuidv4.bytes[9:16]

    return uuid.UUID(bytes=bytes(uuid_bytes))



import tenacity
import httpx

from opik.rest_api.core import api_error as rest_api_error


RETRYABLE_STATUS_CODES = [500, 502, 503, 504]


def _allowed_to_retry(exception: Exception) -> bool:
    if isinstance(
        exception,
        (
            httpx.RemoteProtocolError,  # handle retries for expired connections
            httpx.ConnectError,
            httpx.TimeoutException,
        ),
    ):
        return True

    if isinstance(exception, rest_api_error.ApiError):
        if exception.status_code in RETRYABLE_STATUS_CODES:
            return True

    return False


opik_rest_retry = tenacity.retry(
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_exponential(multiplier=3, min=5, max=15),  # 5, 6, 12
    retry=tenacity.retry_if_exception(_allowed_to_retry),
)



import pydantic
from typing import Union, Dict, Any, Optional
from . import (
    openai_chat_completions_usage,
    google_usage,
    anthropic_usage,
    unknown_usage,
    bedrock_usage,
    openai_responses_usage,
)
from opik import dict_utils

ProviderUsage = Union[
    openai_chat_completions_usage.OpenAICompletionsUsage,
    google_usage.GoogleGeminiUsage,
    anthropic_usage.AnthropicUsage,
    bedrock_usage.BedrockUsage,
    openai_responses_usage.OpenAIResponsesUsage,
    unknown_usage.UnknownUsage,
]


class OpikUsage(pydantic.BaseModel):
    """
        A class used to convert different formats of token usage dictionaries
    into format supported by Opik ecosystem.

        `from_PROVIDER_usage_dict methods` methods are used to parse original provider's token
    usage dicts and calculate openai-formatted extra key-value pairs (that can later be used on the FE and BE sides).
    """

    completion_tokens: Optional[int] = None
    prompt_tokens: Optional[int] = None
    total_tokens: Optional[int] = None

    provider_usage: ProviderUsage

    def to_backend_compatible_full_usage_dict(self) -> Dict[str, int]:
        """
        Returns usage dictionary in backend compatible format:
            * flattened, original usage keys have `original_usage.` prefix
            * only integer values
            * if available, adds openai-formatted keys to the result dict
                so that they can be used on the BE and FE sides.
        """
        short_openai_like_usage: Dict[str, int] = dict_utils.keep_only_values_of_type(
            {
                "completion_tokens": self.completion_tokens,
                "prompt_tokens": self.prompt_tokens,
                "total_tokens": self.total_tokens,
            },
            value_type=int,
        )

        provider_usage: Dict[str, int] = (
            self.provider_usage.to_backend_compatible_flat_dict(
                parent_key_prefix="original_usage"
            )
        )

        return {**short_openai_like_usage, **provider_usage}

    @classmethod
    def from_unknown_usage_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = unknown_usage.UnknownUsage.from_original_usage_dict(usage)

        return cls(provider_usage=provider_usage)

    @classmethod
    def from_openai_completions_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = openai_chat_completions_usage.OpenAICompletionsUsage.from_original_usage_dict(
            usage
        )

        return cls(
            completion_tokens=provider_usage.completion_tokens,
            prompt_tokens=provider_usage.prompt_tokens,
            total_tokens=provider_usage.total_tokens,
            provider_usage=provider_usage,
        )

    @classmethod
    def from_google_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = google_usage.GoogleGeminiUsage.from_original_usage_dict(usage)

        return cls(
            completion_tokens=provider_usage.candidates_token_count,
            prompt_tokens=provider_usage.prompt_token_count,
            total_tokens=provider_usage.total_token_count,
            provider_usage=provider_usage,
        )

    @classmethod
    def from_anthropic_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = anthropic_usage.AnthropicUsage.from_original_usage_dict(usage)

        prompt_tokens = provider_usage.input_tokens + (
            provider_usage.cache_read_input_tokens
            if provider_usage.cache_read_input_tokens is not None
            else 0
        )
        completion_tokens = provider_usage.output_tokens + (
            provider_usage.cache_creation_input_tokens
            if provider_usage.cache_creation_input_tokens is not None
            else 0
        )
        total_tokens = prompt_tokens + completion_tokens

        return cls(
            completion_tokens=completion_tokens,
            prompt_tokens=prompt_tokens,
            total_tokens=total_tokens,
            provider_usage=provider_usage,
        )

    @classmethod
    def from_bedrock_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = bedrock_usage.BedrockUsage.from_original_usage_dict(usage)

        prompt_tokens = provider_usage.inputTokens
        completion_tokens = provider_usage.outputTokens

        total_tokens = prompt_tokens + completion_tokens

        return cls(
            completion_tokens=completion_tokens,
            prompt_tokens=prompt_tokens,
            total_tokens=total_tokens,
            provider_usage=provider_usage,
        )

    @classmethod
    def from_openai_responses_dict(cls, usage: Dict[str, Any]) -> "OpikUsage":
        provider_usage = (
            openai_responses_usage.OpenAIResponsesUsage.from_original_usage_dict(usage)
        )

        return cls(
            completion_tokens=provider_usage.output_tokens,
            prompt_tokens=provider_usage.input_tokens,
            total_tokens=provider_usage.total_tokens,
            provider_usage=provider_usage,
        )



from typing import Set, List, TYPE_CHECKING

if TYPE_CHECKING:
    from opik.guardrails import schemas


class OpikException(Exception):
    pass


class DatasetItemUpdateOperationRequiresItemId(OpikException):
    pass


class ContextExtractorNotSet(OpikException):
    pass


class ConfigurationError(OpikException):
    pass


class ScoreMethodMissingArguments(OpikException):
    pass


class MetricComputationError(OpikException):
    """Exception raised when a metric cannot be computed."""

    pass


class JSONParsingError(OpikException):
    """Exception raised when we fail to parse an LLM response to a dictionary"""

    pass


class PromptPlaceholdersDontMatchFormatArguments(OpikException):
    def __init__(self, prompt_placeholders: Set[str], format_arguments: Set[str]):
        self.prompt_placeholders = prompt_placeholders
        self.format_arguments = format_arguments
        self.symmetric_difference = prompt_placeholders.symmetric_difference(
            format_arguments
        )

    def __str__(self) -> str:
        return (
            f"The `prompt.format(**kwargs)` arguments must exactly match the prompt placeholders. "
            f"Prompt placeholders: {list(self.prompt_placeholders)}. "
            f"Format arguments: {list(self.format_arguments)}. "
            f"Difference: {list(self.symmetric_difference)}. "
        )


class ExperimentNotFound(OpikException):
    pass


class DatasetNotFound(OpikException):
    pass


class GuardrailValidationFailed(OpikException):
    """Exception raised when a guardrail validation fails."""

    def __init__(
        self,
        message: str,
        validation_results: List["schemas.ValidationResult"],
        failed_validations: List["schemas.ValidationResult"],
    ):
        self.message = message
        self.validation_results = validation_results
        self.failed_validations = failed_validations
        super().__init__(message)

    def __str__(self) -> str:
        return f"{self.message}. Failed validations: {self.failed_validations}\n"



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..types.bi_information_response import BiInformationResponse
from ..types.spans_count_response import SpansCountResponse
from ..types.trace_count_response import TraceCountResponse


class RawSystemUsageClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def get_dataset_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[BiInformationResponse]:
        """
        Get datasets information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[BiInformationResponse]
            Datasets BiInformationResponse resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-datasets",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_experiment_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[BiInformationResponse]:
        """
        Get experiments information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[BiInformationResponse]
            Experiments BiInformationResponse resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-experiments",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_spans_count_for_workspaces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[SpansCountResponse]:
        """
        Get spans count on previous day for all available workspaces

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[SpansCountResponse]
            SpanCountResponse resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/internal/usage/workspace-span-counts",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpansCountResponse,
                    parse_obj_as(
                        type_=SpansCountResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_traces_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[BiInformationResponse]:
        """
        Get traces information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[BiInformationResponse]
            Traces BiInformationResponse resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-traces",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_traces_count_for_workspaces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[TraceCountResponse]:
        """
        Get traces count on previous day for all available workspaces

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TraceCountResponse]
            TraceCountResponse resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/internal/usage/workspace-trace-counts",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TraceCountResponse,
                    parse_obj_as(
                        type_=TraceCountResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawSystemUsageClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def get_dataset_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[BiInformationResponse]:
        """
        Get datasets information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[BiInformationResponse]
            Datasets BiInformationResponse resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-datasets",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_experiment_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[BiInformationResponse]:
        """
        Get experiments information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[BiInformationResponse]
            Experiments BiInformationResponse resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-experiments",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_spans_count_for_workspaces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[SpansCountResponse]:
        """
        Get spans count on previous day for all available workspaces

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[SpansCountResponse]
            SpanCountResponse resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/internal/usage/workspace-span-counts",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpansCountResponse,
                    parse_obj_as(
                        type_=SpansCountResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_traces_bi_info(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[BiInformationResponse]:
        """
        Get traces information for BI events per user per workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[BiInformationResponse]
            Traces BiInformationResponse resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/internal/usage/bi-traces",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    BiInformationResponse,
                    parse_obj_as(
                        type_=BiInformationResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_traces_count_for_workspaces(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[TraceCountResponse]:
        """
        Get traces count on previous day for all available workspaces

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TraceCountResponse]
            TraceCountResponse resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/internal/usage/workspace-trace-counts",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TraceCountResponse,
                    parse_obj_as(
                        type_=TraceCountResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import datetime
import logging
from typing import Any, Dict, List, Optional, Union

from opik import datetime_helpers, id_helpers, llm_usage
from opik.message_processing import messages, streamer

from opik.types import (
    DistributedTraceHeadersDict,
    ErrorInfoDict,
    LLMProvider,
    SpanType,
)
from .. import constants, validation_helpers, helpers

LOGGER = logging.getLogger(__name__)


class Span:
    def __init__(
        self,
        id: str,
        trace_id: str,
        project_name: str,
        message_streamer: streamer.Streamer,
        parent_span_id: Optional[str] = None,
    ):
        """
        A Span object. This object should not be created directly, instead use the `span` method of a Trace (:func:`opik.Opik.span`) or another Span (:meth:`opik.Span.span`).
        """
        self.id = id
        self.trace_id = trace_id
        self.parent_span_id = parent_span_id
        self._streamer = message_streamer
        self._project_name = project_name

    def end(
        self,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None,
        model: Optional[str] = None,
        provider: Optional[Union[LLMProvider, str]] = None,
        error_info: Optional[ErrorInfoDict] = None,
        total_cost: Optional[float] = None,
    ) -> None:
        """
        End the span and update its attributes.

        This method is similar to the `update` method, but it automatically computes
        the end time if not provided.

        Args:
            end_time: The end time of the span. If not provided, the current time will be used.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage data for the span. In order for input, output and total tokens to be visible in the UI,
                the usage must contain OpenAI-formatted keys (they can be passed additionaly to original usage on the top level of the dict):  prompt_tokens, completion_tokens and total_tokens.
                If OpenAI-formatted keys were not found, Opik will try to calculate them automatically if the usage
                format is recognized (you can see which provider's formats are recognized in opik.LLMProvider enum), but it is not guaranteed.
            model: The name of LLM.
            provider: The provider of LLM. You can find providers officially supported by Opik for cost tracking
                in `opik.LLMProvider` enum. If your provider is not here, please open an issue in our github - https://github.com/comet-ml/opik.
                If your provider not in the list, you can still specify it but the cost tracking will not be available
            error_info: The dictionary with error information (typically used when the span function has failed).
            total_cost: The cost of the span in USD. This value takes priority over the cost calculated by Opik from the usage.

        Returns:
            None
        """
        end_time = (
            end_time if end_time is not None else datetime_helpers.local_timestamp()
        )

        self.update(
            end_time=end_time,
            metadata=metadata,
            input=input,
            output=output,
            tags=tags,
            usage=usage,
            model=model,
            provider=provider,
            error_info=error_info,
            total_cost=total_cost,
        )

    def update(
        self,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None,
        model: Optional[str] = None,
        provider: Optional[Union[LLMProvider, str]] = None,
        error_info: Optional[ErrorInfoDict] = None,
        total_cost: Optional[float] = None,
    ) -> None:
        """
        Update the span attributes.

        Args:
            end_time: The end time of the span.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage data for the span. In order for input, output and total tokens to be visible in the UI,
                the usage must contain OpenAI-formatted keys (they can be passed additionaly to original usage on the top level of the dict):  prompt_tokens, completion_tokens and total_tokens.
                If OpenAI-formatted keys were not found, Opik will try to calculate them automatically if the usage
                format is recognized (you can see which provider's formats are recognized in opik.LLMProvider enum), but it is not guaranteed.
            model: The name of LLM.
            provider: The provider of LLM. You can find providers officially supported by Opik for cost tracking
                in `opik.LLMProvider` enum. If your provider is not here, please open an issue in our github - https://github.com/comet-ml/opik.
                If your provider not in the list, you can still specify it but the cost tracking will not be available
            error_info: The dictionary with error information (typically used when the span function has failed).
            total_cost: The cost of the span in USD. This value takes priority over the cost calculated by Opik from the usage.

        Returns:
            None
        """
        backend_compatible_usage = validation_helpers.validate_and_parse_usage(
            usage=usage,
            logger=LOGGER,
            provider=provider,
        )

        if backend_compatible_usage is not None:
            metadata = helpers.add_usage_to_metadata(usage=usage, metadata=metadata)

        end_span_message = messages.UpdateSpanMessage(
            span_id=self.id,
            trace_id=self.trace_id,
            parent_span_id=self.parent_span_id,
            project_name=self._project_name,
            end_time=end_time,
            metadata=metadata,
            input=input,
            output=output,
            tags=tags,
            usage=backend_compatible_usage,
            model=model,
            provider=provider,
            error_info=error_info,
            total_cost=total_cost,
        )
        self._streamer.put(end_span_message)

    def span(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        type: SpanType = "general",
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None,
        model: Optional[str] = None,
        provider: LLMProvider = LLMProvider.OPENAI,
        error_info: Optional[ErrorInfoDict] = None,
        total_cost: Optional[float] = None,
    ) -> "Span":
        """
        Create a new child span within the current span.

        Args:
            id: The ID of the span, should be in UUIDv7 format. If not provided, a new ID will be generated.
            name: The name of the span.
            type: The type of the span. Defaults to "general".
            start_time: The start time of the span. If not provided, current time will be used.
            end_time: The end time of the span.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage data for the span. In order for input, output and total tokens to be visible in the UI,
                the usage must contain OpenAI-formatted keys (they can be passed additionaly to original usage on the top level of the dict):  prompt_tokens, completion_tokens and total_tokens.
                If OpenAI-formatted keys were not found, Opik will try to calculate them automatically if the usage
                format is recognized (you can see which provider's formats are recognized in opik.LLMProvider enum), but it is not guaranteed.
            model: The name of LLM (in this case `type` parameter should be == `llm`)
            provider: The provider of LLM. You can find providers officially supported by Opik for cost tracking
                in `opik.LLMProvider` enum. If your provider is not here, please open an issue in our github - https://github.com/comet-ml/opik.
                If your provider not in the list, you can still specify it but the cost tracking will not be available
            error_info: The dictionary with error information (typically used when the span function has failed).
            total_cost: The cost of the span in USD. This value takes priority over the cost calculated by Opik from the usage.

        Returns:
            Span: The created child span object.
        """
        span_id = id if id is not None else id_helpers.generate_id()
        start_time = (
            start_time if start_time is not None else datetime_helpers.local_timestamp()
        )
        backend_compatible_usage = validation_helpers.validate_and_parse_usage(
            usage=usage,
            logger=LOGGER,
            provider=provider,
        )

        if backend_compatible_usage is not None:
            metadata = helpers.add_usage_to_metadata(usage=usage, metadata=metadata)

        create_span_message = messages.CreateSpanMessage(
            span_id=span_id,
            trace_id=self.trace_id,
            project_name=self._project_name,
            parent_span_id=self.id,
            name=name,
            type=type,
            start_time=start_time,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            tags=tags,
            usage=backend_compatible_usage,
            model=model,
            provider=provider,
            error_info=error_info,
            total_cost=total_cost,
        )
        self._streamer.put(create_span_message)

        return Span(
            id=span_id,
            parent_span_id=self.id,
            trace_id=self.trace_id,
            message_streamer=self._streamer,
            project_name=self._project_name,
        )

    def log_feedback_score(
        self,
        name: str,
        value: float,
        category_name: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None:
        """
        Log a feedback score for the span.

        Args:
            name: The name of the feedback score.
            value: The value of the feedback score.
            category_name: The category name for the feedback score.
            reason: The reason for the feedback score.

        Returns:
            None
        """
        add_span_feedback_batch_message = messages.AddSpanFeedbackScoresBatchMessage(
            batch=[
                messages.FeedbackScoreMessage(
                    id=self.id,
                    name=name,
                    value=value,
                    category_name=category_name,
                    reason=reason,
                    source=constants.FEEDBACK_SCORE_SOURCE_SDK,
                    project_name=self._project_name,
                )
            ],
        )

        self._streamer.put(add_span_feedback_batch_message)

    def get_distributed_trace_headers(self) -> DistributedTraceHeadersDict:
        """
        Returns headers dictionary to be passed into tracked
        function on remote node.
        """
        return {"opik_parent_span_id": self.id, "opik_trace_id": self.trace_id}



from typing import Optional

from google import genai

from . import (
    generate_content_decorator,
    generations_aggregators,
    stream_wrappers,
    encoder_extension,
)


def track_genai(
    client: genai.Client, project_name: Optional[str] = None
) -> genai.Client:
    """
    Adds Opik tracking to an genai.Client.

    Tracks calls to:
    * client.models.generate_content
    * client.models.generate_content_stream
    * client.aio.models.generate_content
    * client.aio.models.generate_content_stream

    Can be used within other Opik-tracked functions.

    Args:
        client: An instance of genai.Client.
        project_name: The name of the project to log data.

    Returns:
        The modified genai.Client with Opik tracking enabled.
    """
    if hasattr(client, "opik_tracked"):
        return client
    encoder_extension.register()

    client.opik_tracked = True

    provider = "google_vertexai" if client.vertexai else "google_ai"

    decorator_factory = generate_content_decorator.GenerateContentTrackDecorator(
        provider=provider
    )

    client.models.generate_content = decorator_factory.track(
        name="generate_content",
        type="llm",
        project_name=project_name,
    )(client.models.generate_content)

    # This conversion is needed so that @track decorator stopped treating generate_content_stream
    # as generator function, as they are tracked slightly differently from what we need here.
    # The user still gets their Iterator object in return.
    client.models.generate_content_stream = (
        stream_wrappers.generator_function_to_normal_function(
            client.models.generate_content_stream
        )
    )
    client.models.generate_content_stream = decorator_factory.track(
        name="generate_content_stream",
        type="llm",
        project_name=project_name,
        generations_aggregator=generations_aggregators.aggregate_response_content_items,
    )(client.models.generate_content_stream)

    client.aio.models.generate_content = decorator_factory.track(
        name="async_generate_content",
        type="llm",
        project_name=project_name,
    )(client.aio.models.generate_content)

    # No need to perform a similar conversion as for the synchronous method, because async version of generate_content_stream
    # is already a wrapper around generator function that works similarly to our helper function
    client.aio.models.generate_content_stream = decorator_factory.track(
        name="async_generate_content_stream",
        type="llm",
        project_name=project_name,
        generations_aggregator=generations_aggregators.aggregate_response_content_items,
    )(client.aio.models.generate_content_stream)

    return client



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.feedback_create import FeedbackCreate
from ..types.feedback_definition_page_public import FeedbackDefinitionPagePublic
from ..types.feedback_public import FeedbackPublic
from ..types.feedback_update import FeedbackUpdate
from .raw_client import AsyncRawFeedbackDefinitionsClient, RawFeedbackDefinitionsClient
from .types.find_feedback_definitions_request_type import FindFeedbackDefinitionsRequestType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class FeedbackDefinitionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawFeedbackDefinitionsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawFeedbackDefinitionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawFeedbackDefinitionsClient
        """
        return self._raw_client

    def find_feedback_definitions(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackDefinitionsRequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FeedbackDefinitionPagePublic:
        """
        Find Feedback definitions

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        type : typing.Optional[FindFeedbackDefinitionsRequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackDefinitionPagePublic
            Feedback definitions resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.find_feedback_definitions()
        """
        _response = self._raw_client.find_feedback_definitions(
            page=page, size=size, name=name, type=type, request_options=request_options
        )
        return _response.data

    def create_feedback_definition(
        self, *, request: FeedbackCreate, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Get feedback definition

        Parameters
        ----------
        request : FeedbackCreate

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import FeedbackCreate_Numerical
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.create_feedback_definition(request=FeedbackCreate_Numerical(), )
        """
        _response = self._raw_client.create_feedback_definition(request=request, request_options=request_options)
        return _response.data

    def get_feedback_definition_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackPublic:
        """
        Get feedback definition by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackPublic
            Feedback definition resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.get_feedback_definition_by_id(id='id', )
        """
        _response = self._raw_client.get_feedback_definition_by_id(id, request_options=request_options)
        return _response.data

    def update_feedback_definition(
        self, id: str, *, request: FeedbackUpdate, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Update feedback definition by id

        Parameters
        ----------
        id : str

        request : FeedbackUpdate

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import FeedbackUpdate_Numerical
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.update_feedback_definition(id='id', request=FeedbackUpdate_Numerical(), )
        """
        _response = self._raw_client.update_feedback_definition(id, request=request, request_options=request_options)
        return _response.data

    def delete_feedback_definition_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete feedback definition by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.delete_feedback_definition_by_id(id='id', )
        """
        _response = self._raw_client.delete_feedback_definition_by_id(id, request_options=request_options)
        return _response.data

    def delete_feedback_definitions_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete feedback definitions batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.feedback_definitions.delete_feedback_definitions_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_feedback_definitions_batch(ids=ids, request_options=request_options)
        return _response.data


class AsyncFeedbackDefinitionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawFeedbackDefinitionsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawFeedbackDefinitionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawFeedbackDefinitionsClient
        """
        return self._raw_client

    async def find_feedback_definitions(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackDefinitionsRequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FeedbackDefinitionPagePublic:
        """
        Find Feedback definitions

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        type : typing.Optional[FindFeedbackDefinitionsRequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackDefinitionPagePublic
            Feedback definitions resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.find_feedback_definitions()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_feedback_definitions(
            page=page, size=size, name=name, type=type, request_options=request_options
        )
        return _response.data

    async def create_feedback_definition(
        self, *, request: FeedbackCreate, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Get feedback definition

        Parameters
        ----------
        request : FeedbackCreate

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import FeedbackCreate_Numerical
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.create_feedback_definition(request=FeedbackCreate_Numerical(), )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_feedback_definition(request=request, request_options=request_options)
        return _response.data

    async def get_feedback_definition_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackPublic:
        """
        Get feedback definition by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackPublic
            Feedback definition resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.get_feedback_definition_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_feedback_definition_by_id(id, request_options=request_options)
        return _response.data

    async def update_feedback_definition(
        self, id: str, *, request: FeedbackUpdate, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Update feedback definition by id

        Parameters
        ----------
        id : str

        request : FeedbackUpdate

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import FeedbackUpdate_Numerical
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.update_feedback_definition(id='id', request=FeedbackUpdate_Numerical(), )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_feedback_definition(
            id, request=request, request_options=request_options
        )
        return _response.data

    async def delete_feedback_definition_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete feedback definition by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.delete_feedback_definition_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_feedback_definition_by_id(id, request_options=request_options)
        return _response.data

    async def delete_feedback_definitions_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete feedback definitions batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.feedback_definitions.delete_feedback_definitions_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_feedback_definitions_batch(ids=ids, request_options=request_options)
        return _response.data



import gzip
from typing import Optional, Dict, Any, Union, Iterable, AsyncIterable, Mapping
import httpx
import os
import json as jsonlib

from . import hooks, package_version
import platform


CABundlePath = str

KEEPALIVE_EXPIRY_SECONDS = 10
CONNECT_TIMEOUT_SECONDS = 20
READ_TIMEOUT_SECONDS = 100
WRITE_TIMEOUT_SECONDS = 100
POOL_TIMEOUT_SECONDS = 20


def get(
    workspace: Optional[str],
    api_key: Optional[str],
    check_tls_certificate: bool,
    compress_json_requests: bool,
) -> httpx.Client:
    limits = httpx.Limits(keepalive_expiry=KEEPALIVE_EXPIRY_SECONDS)

    verify: Union[bool, CABundlePath] = (
        os.environ["SSL_CERT_FILE"]
        if check_tls_certificate is True and "SSL_CERT_FILE" in os.environ
        else check_tls_certificate
    )

    timeout = httpx.Timeout(
        connect=CONNECT_TIMEOUT_SECONDS,
        read=READ_TIMEOUT_SECONDS,
        write=WRITE_TIMEOUT_SECONDS,
        pool=POOL_TIMEOUT_SECONDS,
    )

    client = OpikHttpxClient(
        compress_json_requests=compress_json_requests,
        limits=limits,
        verify=verify,
        timeout=timeout,
        follow_redirects=True,
    )

    headers = _prepare_headers(workspace=workspace, api_key=api_key)
    client.headers.update(headers)

    hooks.run_httpx_client_hooks(client)

    return client


def _prepare_headers(
    workspace: Optional[str], api_key: Optional[str]
) -> Dict[str, Any]:
    result = {
        "X-OPIK-DEBUG-SDK-VERSION": package_version.VERSION,
        "X-OPIK-DEBUG-PY-VERSION": platform.python_version(),
    }

    if workspace is not None:
        result["Comet-Workspace"] = workspace

    if api_key is not None:
        result["Authorization"] = api_key

    return result


class OpikHttpxClient(httpx.Client):
    def __init__(self, compress_json_requests: bool = True, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.compress_json_requests = compress_json_requests

    def build_request(
        self,
        method: str,
        url: Union[httpx.URL, str],
        *,
        content: Optional[
            Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]
        ] = None,
        data: Optional[Mapping[str, Any]] = None,
        files: Any = None,
        json: Any = None,
        params: Any = None,
        headers: Any = None,
        cookies: Any = None,
        timeout: Any = httpx.USE_CLIENT_DEFAULT,
        extensions: Any = None,
    ) -> httpx.Request:
        # we override this method to allow compression of JSON requests that is handled
        # by httpx.Client.request() as well as by httpx.Client.stream() (both used in the OPIK)
        if self.compress_json_requests:
            if method in ("POST", "PUT", "PATCH") and json is not None:
                json_data = jsonlib.dumps(json).encode("utf-8")
                content = gzip.compress(json_data)
                json = None
                if headers is None:
                    headers = {}
                headers["Content-Length"] = str(len(content))
                headers["Content-Encoding"] = "gzip"
                if "content-type" not in headers:
                    # to avoid having it in headers two times with different cases in keys (e.g., streaming operations)
                    headers["Content-Type"] = "application/json;charset=utf-8"

        return super().build_request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            extensions=extensions,
        )



from typing import List, Optional


class ValidationResult:
    def __init__(self, failed: bool, failure_reasons: Optional[List[str]] = None):
        self._failed = failed
        self._failure_reasons: List[str] = (
            failure_reasons if failure_reasons is not None else []
        )

    def failed(self) -> bool:
        return self._failed

    def ok(self) -> bool:
        return not self._failed

    @property
    def failure_reasons(self) -> List[str]:
        return self._failure_reasons

    def __bool__(self) -> bool:
        return not self._failed

    def __str__(self) -> str:
        if not self._failed:
            return "OK"

        return f"Failed: {self._failure_reasons}"



import functools
import logging
from typing import List, Optional

from opik.message_processing.batching import sequence_splitter
from opik.rest_api import client as rest_api_client
from opik.rest_api.types import experiment_item as rest_experiment_item
from . import experiment_item
from .. import constants, helpers
from ...api_objects.prompt import Prompt

LOGGER = logging.getLogger(__name__)


class Experiment:
    def __init__(
        self,
        id: str,
        name: Optional[str],
        dataset_name: str,
        rest_client: rest_api_client.OpikApi,
        prompts: Optional[List[Prompt]] = None,
    ) -> None:
        self._id = id
        self._name = name
        self._dataset_name = dataset_name
        self._rest_client = rest_client
        self._prompts = prompts

    @property
    def id(self) -> str:
        return self._id

    @functools.cached_property
    def dataset_id(self) -> str:
        return self._rest_client.datasets.get_dataset_by_identifier(
            dataset_name=self._dataset_name
        ).id

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @functools.cached_property
    def name(self) -> str:
        if self._name is not None:
            return self._name

        return self._rest_client.experiments.get_experiment_by_id(id=self.id).name

    def insert(
        self,
        experiment_items_references: List[experiment_item.ExperimentItemReferences],
    ) -> None:
        """
        Creates a new experiment item by linking the existing trace and dataset item.

        Args:
            experiment_items_references: The list of ExperimentItemReferences objects, containing
                trace id and dataset item id to link together into experiment item.

        Returns:
            None
        """
        rest_experiment_items = [
            rest_experiment_item.ExperimentItem(
                id=helpers.generate_id(),
                experiment_id=self.id,
                dataset_item_id=item.dataset_item_id,
                trace_id=item.trace_id,
            )
            for item in experiment_items_references
        ]

        batches = sequence_splitter.split_into_batches(
            rest_experiment_items, max_length=constants.EXPERIMENT_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Sending experiment items batch: %s", batch)
            self._rest_client.experiments.create_experiment_items(
                experiment_items=batch,
            )
            LOGGER.debug("Sent experiment items batch of size %d", len(batch))

    def get_items(self) -> List[experiment_item.ExperimentItemContent]:
        """
        Returns:
            List[ExperimentItemContent]: the list with contents of existing experiment items.
        """
        result: List[experiment_item.ExperimentItemContent] = []

        page = 0

        while True:  # TODO: refactor this logic when backend implements a proper streaming endpoint
            page += 1
            dataset_items_page = (
                self._rest_client.datasets.find_dataset_items_with_experiment_items(
                    id=self.dataset_id,
                    experiment_ids=f'["{self.id}"]',
                    page=page,
                    size=100,
                )
            )
            if len(dataset_items_page.content) == 0:
                break

            for dataset_item in dataset_items_page.content:
                rest_experiment_item_compare = dataset_item.experiment_items[0]
                content = experiment_item.ExperimentItemContent.from_rest_experiment_item_compare(
                    value=rest_experiment_item_compare
                )
                result.append(content)

        return result



G_EVAL_COT_TEMPLATE = """
*** TASK:
Based on the following task description and evaluation criteria,
generate a detailed Chain of Thought (CoT) that outlines the necessary Evaluation Steps
to assess the solution. The CoT should clarify the reasoning process for each step of evaluation.

*** INPUT:

TASK INTRODUCTION:
{task_introduction}

EVALUATION CRITERIA:
{evaluation_criteria}

FINAL SCORE:
IF THE USER'S SCALE IS DIFFERENT FROM THE 0 TO 10 RANGE, RECALCULATE THE VALUE USING THIS SCALE.
SCORE VALUE MUST BE AN INTEGER.
"""


G_EVAL_QUERY_TEMPLATE = """
*** TASK INTRODUCTION:
{task_introduction}

*** EVALUATION CRITERIA:
{evaluation_criteria}

{chain_of_thought}

*** INPUT:
{input}

*** OUTPUT:
Return the output in a JSON format with the keys "score" and "reason".
"""



from typing import List, TypedDict


class FewShotExampleWithContextAnswerRelevance(TypedDict):
    title: str
    input: str
    output: str
    context: List[str]
    answer_relevance_score: float
    reason: str


class FewShotExampleNoContextAnswerRelevance(TypedDict):
    title: str
    input: str
    output: str
    answer_relevance_score: float
    reason: str


FEW_SHOT_EXAMPLES_WITH_CONTEXT: List[FewShotExampleWithContextAnswerRelevance] = [
    {
        "title": "Low Relevance Score",
        "input": "What's the capital of France?",
        "output": "The Eiffel Tower is a famous landmark.",
        "context": [
            "France is a country in Europe.",
            "Paris is known for its iconic Eiffel Tower.",
        ],
        "answer_relevance_score": 0.2,
        "reason": "The answer provides information about the Eiffel Tower, which is related to France, but fails to address the specific question about the capital city. It doesn't directly answer the user's query, resulting in low relevance.",
    },
    {
        "title": "Medium Relevance Score",
        "input": "What's the capital of France?",
        "output": "France has many beautiful cities, including Paris.",
        "context": [
            "France is a country in Europe.",
            "Paris is the capital and largest city of France.",
        ],
        "answer_relevance_score": 0.6,
        "reason": "The answer mentions Paris, which is the correct capital, but it's presented as just one of many cities rather than explicitly stating it's the capital. The response is partially relevant but lacks directness in addressing the specific question.",
    },
    {
        "title": "High Relevance Score",
        "input": "What's the capital of France?",
        "output": "The capital of France is Paris, a city known for its iconic Eiffel Tower.",
        "context": [
            "France is a country in Europe.",
            "Paris is the capital and largest city of France.",
        ],
        "answer_relevance_score": 0.9,
        "reason": "The answer directly and correctly identifies Paris as the capital of France, which is highly relevant to the user's question. It also provides additional context about the Eiffel Tower, which aligns with the provided context. The response is comprehensive and relevant, though slightly more detailed than necessary, preventing a perfect score.",
    },
]

FEW_SHOT_EXAMPLES_NO_CONTEXT: List[FewShotExampleNoContextAnswerRelevance] = [
    {
        "title": "Low Relevance Score",
        "input": "What's the capital of France?",
        "output": "The Eiffel Tower is a famous landmark.",
        "answer_relevance_score": 0.2,
        "reason": "The answer provides information about the Eiffel Tower, which is related to France, but fails to address the specific question about the capital city. It doesn't directly answer the user's query, resulting in low relevance.",
    },
    {
        "title": "Medium Relevance Score",
        "input": "What's the capital of France?",
        "output": "France has many beautiful cities, including Paris.",
        "answer_relevance_score": 0.6,
        "reason": "The answer mentions Paris, which is the correct capital, but it's presented as just one of many cities rather than explicitly stating it's the capital. The response is partially relevant but lacks directness in addressing the specific question.",
    },
    {
        "title": "High Relevance Score",
        "input": "What's the capital of France?",
        "output": "The capital of France is Paris, a city known for its iconic Eiffel Tower.",
        "answer_relevance_score": 0.9,
        "reason": "The answer directly and correctly identifies Paris as the capital of France, which is highly relevant to the user's question. It also provides additional context about the Eiffel Tower, which aligns with the provided context. The response is comprehensive and relevant, though slightly more detailed than necessary, preventing a perfect score.",
    },
]


def generate_query_with_context(
    input: str,
    output: str,
    context: List[str],
    few_shot_examples: List[FewShotExampleWithContextAnswerRelevance],
) -> str:
    examples_str = "\n\n".join(
        [
            f"#### Example {i + 1}: {example['title']}\n\n"
            f'- **Input:** "{example["input"]}"\n'
            f'- **Output:** "{example["output"]}"\n'
            f"- **Context:** {example['context']}\n"
            f"- **Result:**\n"
            f"  ```json\n"
            f"  {{\n"
            f'    "answer_relevance_score": {example["answer_relevance_score"]},\n'
            f'    "reason": "{example["reason"]}"\n'
            f"  }}\n"
            f"  ```"
            for i, example in enumerate(few_shot_examples)
        ]
    )

    return f"""
        YOU ARE AN EXPERT IN NLP EVALUATION METRICS, SPECIALLY TRAINED TO ASSESS ANSWER RELEVANCE IN RESPONSES
        PROVIDED BY LANGUAGE MODELS. YOUR TASK IS TO EVALUATE THE RELEVANCE OF A GIVEN ANSWER FROM
        ANOTHER LLM BASED ON THE USER'S INPUT AND CONTEXT PROVIDED.

        ###INSTRUCTIONS###

        - YOU MUST ANALYZE THE GIVEN CONTEXT AND USER INPUT TO DETERMINE THE MOST RELEVANT RESPONSE.
        - EVALUATE THE ANSWER FROM THE OTHER LLM BASED ON ITS ALIGNMENT WITH THE USER'S QUERY AND THE CONTEXT.
        - ASSIGN A RELEVANCE SCORE BETWEEN 0.0 (COMPLETELY IRRELEVANT) AND 1.0 (HIGHLY RELEVANT).
        - RETURN THE RESULT AS A JSON OBJECT, INCLUDING THE SCORE AND A BRIEF EXPLANATION OF THE RATING.

        ###CHAIN OF THOUGHTS###

        1. **Understanding the Context and Input:**
           1.1. READ AND COMPREHEND THE CONTEXT PROVIDED.
           1.2. IDENTIFY THE KEY POINTS OR QUESTIONS IN THE USER'S INPUT THAT THE ANSWER SHOULD ADDRESS.

        2. **Evaluating the Answer:**
           2.1. COMPARE THE CONTENT OF THE ANSWER TO THE CONTEXT AND USER INPUT.
           2.2. DETERMINE WHETHER THE ANSWER DIRECTLY ADDRESSES THE USER'S QUERY OR PROVIDES RELEVANT INFORMATION.
           2.3. CONSIDER ANY EXTRANEOUS OR OFF-TOPIC INFORMATION THAT MAY DECREASE RELEVANCE.

        3. **Assigning a Relevance Score:**
           3.1. ASSIGN A SCORE BASED ON HOW WELL THE ANSWER MATCHES THE USER'S NEEDS AND CONTEXT.
           3.2. JUSTIFY THE SCORE WITH A BRIEF EXPLANATION THAT HIGHLIGHTS THE STRENGTHS OR WEAKNESSES OF THE ANSWER.

        4. **Generating the JSON Output:**
           4.1. FORMAT THE OUTPUT AS A JSON OBJECT WITH A "answer_relevance_score" FIELD AND AN "reason" FIELD.
           4.2. ENSURE THE SCORE IS A FLOATING-POINT NUMBER BETWEEN 0.0 AND 1.0.

        ###WHAT NOT TO DO###

        - DO NOT GIVE A SCORE WITHOUT FULLY ANALYZING BOTH THE CONTEXT AND THE USER INPUT.
        - AVOID SCORES THAT DO NOT MATCH THE EXPLANATION PROVIDED.
        - DO NOT INCLUDE ADDITIONAL FIELDS OR INFORMATION IN THE JSON OUTPUT BEYOND "answer_relevance_score" AND "reason."
        - NEVER ASSIGN A PERFECT SCORE UNLESS THE ANSWER IS FULLY RELEVANT AND FREE OF ANY IRRELEVANT INFORMATION.

        ###EXAMPLE OUTPUT FORMAT###
        {{
          "answer_relevance_score": 0.85,
          "reason": "The answer addresses the user's query about the primary topic but includes some extraneous details that slightly reduce its relevance."
        }}

        ###FEW-SHOT EXAMPLES###

        {examples_str}

        ###INPUTS:###
        ***
        Input:
        {input}

        Output:
        {output}

        Context:
        {context}
        ***
    """


def generate_query_no_context(
    input: str,
    output: str,
    few_shot_examples: List[FewShotExampleNoContextAnswerRelevance],
) -> str:
    examples_str = "\n\n".join(
        [
            f"#### Example {i + 1}: {example['title']}\n\n"
            f'- **Input:** "{example["input"]}"\n'
            f'- **Output:** "{example["output"]}"\n'
            f"- **Result:**\n"
            f"  ```json\n"
            f"  {{\n"
            f'    "answer_relevance_score": {example["answer_relevance_score"]},\n'
            f'    "reason": "{example["reason"]}"\n'
            f"  }}\n"
            f"  ```"
            for i, example in enumerate(few_shot_examples)
        ]
    )

    return f"""
        YOU ARE AN EXPERT IN NLP EVALUATION METRICS, SPECIALLY TRAINED TO ASSESS ANSWER RELEVANCE IN RESPONSES
        PROVIDED BY LANGUAGE MODELS. YOUR TASK IS TO EVALUATE THE RELEVANCE OF A GIVEN ANSWER FROM
        ANOTHER LLM BASED ON THE USER'S INPUT.

        ###INSTRUCTIONS###

        - YOU MUST ANALYZE THE USER INPUT TO DETERMINE THE MOST RELEVANT RESPONSE.
        - EVALUATE THE ANSWER FROM THE OTHER LLM BASED ON ITS ALIGNMENT WITH THE USER'S QUERY.
        - ASSIGN A RELEVANCE SCORE BETWEEN 0.0 (COMPLETELY IRRELEVANT) AND 1.0 (HIGHLY RELEVANT).
        - RETURN THE RESULT AS A JSON OBJECT, INCLUDING THE SCORE AND A BRIEF EXPLANATION OF THE RATING.

        ###CHAIN OF THOUGHTS###

        1. **Understanding the Input:**
           1.1. IDENTIFY THE KEY POINTS OR QUESTIONS IN THE USER'S INPUT THAT THE ANSWER SHOULD ADDRESS.

        2. **Evaluating the Answer:**
           2.1. COMPARE THE CONTENT OF THE ANSWER TO THE USER INPUT.
           2.2. DETERMINE WHETHER THE ANSWER DIRECTLY ADDRESSES THE USER'S QUERY OR PROVIDES RELEVANT INFORMATION.
           2.3. CONSIDER ANY EXTRANEOUS OR OFF-TOPIC INFORMATION THAT MAY DECREASE RELEVANCE.

        3. **Assigning a Relevance Score:**
           3.1. ASSIGN A SCORE BASED ON HOW WELL THE ANSWER MATCHES THE USER'S NEEDS
           3.2. JUSTIFY THE SCORE WITH A BRIEF EXPLANATION THAT HIGHLIGHTS THE STRENGTHS OR WEAKNESSES OF THE ANSWER.

        4. **Generating the JSON Output:**
           4.1. FORMAT THE OUTPUT AS A JSON OBJECT WITH A "answer_relevance_score" FIELD AND AN "reason" FIELD.
           4.2. ENSURE THE SCORE IS A FLOATING-POINT NUMBER BETWEEN 0.0 AND 1.0.

        ###WHAT NOT TO DO###

        - DO NOT GIVE A SCORE WITHOUT FULLY ANALYZING THE USER INPUT.
        - AVOID SCORES THAT DO NOT MATCH THE EXPLANATION PROVIDED.
        - DO NOT INCLUDE ADDITIONAL FIELDS OR INFORMATION IN THE JSON OUTPUT BEYOND "answer_relevance_score" AND "reason."
        - NEVER ASSIGN A PERFECT SCORE UNLESS THE ANSWER IS FULLY RELEVANT AND FREE OF ANY IRRELEVANT INFORMATION.

        ###EXAMPLE OUTPUT FORMAT###
        {{
          "answer_relevance_score": 0.85,
          "reason": "The answer addresses the user's query about the primary topic but includes some extraneous details that slightly reduce its relevance."
        }}

        ###FEW-SHOT EXAMPLES###

        {examples_str}

        ###INPUTS:###
        ***
        Input:
        {input}

        Output:
        {output}
        ***
    """



import dataclasses
from typing import Any, Dict, Optional
import logging

from agents import tracing

from opik.types import SpanType, LLMProvider
from opik import dict_utils, llm_usage

LOGGER = logging.getLogger(__name__)


def span_type(openai_span_data: tracing.SpanData) -> SpanType:
    if openai_span_data.type in ["function", "guardrail"]:
        return "tool"
    elif openai_span_data.type in ["generation", "response"]:
        return "llm"

    return "general"


def _span_name(openai_span_data: tracing.SpanData) -> str:
    if (
        isinstance(openai_span_data, tracing.AgentSpanData)
        or isinstance(openai_span_data, tracing.FunctionSpanData)
        or isinstance(openai_span_data, tracing.GuardrailSpanData)
        or isinstance(openai_span_data, tracing.CustomSpanData)
    ):
        return openai_span_data.name
    elif isinstance(openai_span_data, tracing.GenerationSpanData):
        return "Generation"
    elif isinstance(openai_span_data, tracing.ResponseSpanData):
        return "Response"
    elif isinstance(openai_span_data, tracing.HandoffSpanData):
        return "Handoff"
    else:
        return "Unknown"


@dataclasses.dataclass
class ParsedSpanData:
    name: str
    type: SpanType
    input: Dict[str, Any]
    output: Dict[str, Any]
    metadata: Dict[str, Any]
    usage: Optional[llm_usage.OpikUsage] = None
    model: Optional[str] = None
    provider: str = LLMProvider.OPENAI


def parse_spandata(openai_span_data: tracing.SpanData) -> ParsedSpanData:
    """
    Parses the given SpanData instance and extracts input, output, and metadata.

    :param spandata: An instance of SpanData or its subclass.
    :return: A ParsedSpanData dataclass containing input, output, and metadata
    """
    content = openai_span_data.export()

    metadata = None
    input_data = None
    output_data = None
    name = _span_name(openai_span_data)
    type = span_type(openai_span_data)

    if openai_span_data.type == "function":
        input_data = openai_span_data.input
        output_data = openai_span_data.output
        del content["input"], content["output"]

    elif openai_span_data.type == "generation":
        input_data = openai_span_data.input
        output_data = openai_span_data.output
        del content["input"], content["output"]

    elif openai_span_data.type == "response":
        return _parse_response_span_content(openai_span_data)

    elif openai_span_data.type == "agent":
        output_data = openai_span_data.output_type

    elif openai_span_data.type == "handoff":
        pass  # No explicit input or output

    elif openai_span_data.type == "custom":
        input_data = openai_span_data.data.get("input")
        output_data = openai_span_data.data.get("output")
        del content["data"]

    elif openai_span_data.type == "guardrail":
        pass  # No explicit input or output

    if input_data is not None and not isinstance(input_data, dict):
        input_data = {"input": input_data}

    if output_data is not None and not isinstance(output_data, dict):
        output_data = {"output": output_data}

    metadata = content

    return ParsedSpanData(
        input=input_data,
        output=output_data,
        metadata=metadata,
        name=name,
        type=type,
    )


def _parse_response_span_content(span_data: tracing.ResponseSpanData) -> ParsedSpanData:
    response = span_data.response
    response_dict = span_data.response.model_dump()
    input = {"input": span_data.input}
    output = {"output": response.output}

    _, metadata = dict_utils.split_dict_by_keys(
        input_dict=response_dict,
        keys=["input", "output"],
    )

    if response.usage is not None:
        opik_usage = llm_usage.try_build_opik_usage_or_log_error(
            provider="_openai_responses",
            usage=response.usage.model_dump(),
            logger=LOGGER,
            error_message="Failed to log usage in openai agent run",
        )
    else:
        opik_usage = None

    return ParsedSpanData(
        name="Response",
        input=input,
        output=output,
        usage=opik_usage,
        type="llm",
        metadata=metadata,
        model=response.model,
    )



import logging
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Tuple, Union

from opik import llm_usage
from opik.types import LLMProvider

if TYPE_CHECKING:
    from langchain_core.tracers.schemas import Run

LOGGER = logging.getLogger(__name__)


def get_llm_usage_info(
    run_dict: Optional[Dict[str, Any]] = None,
) -> llm_usage.LLMUsageInfo:
    if run_dict is None:
        return llm_usage.LLMUsageInfo()

    usage_dict = _try_get_token_usage(run_dict)
    provider, model = _get_provider_and_model(run_dict)

    return llm_usage.LLMUsageInfo(provider=provider, model=model, usage=usage_dict)


def _try_get_token_usage(run_dict: Dict[str, Any]) -> Optional[llm_usage.OpikUsage]:
    try:
        if run_dict["outputs"]["llm_output"] is not None:
            usage_dict = run_dict["outputs"]["llm_output"]["usage"]
        else:
            # Handle the streaming mode
            usage_dict = run_dict["outputs"]["generations"][-1][-1]["message"][
                "kwargs"
            ]["usage_metadata"]

            # Support streaming mode
            if "input_token_details" in usage_dict:
                cache_creation_input_tokens = usage_dict["input_token_details"][
                    "cache_creation"
                ]
                cache_read_input_tokens = usage_dict["input_token_details"][
                    "cache_read"
                ]
            else:
                cache_creation_input_tokens = None
                cache_read_input_tokens = None

            usage_dict = {
                "input_tokens": usage_dict["input_tokens"],
                "output_tokens": usage_dict["output_tokens"],
                "cache_creation_input_tokens": cache_creation_input_tokens,
                "cache_read_input_tokens": cache_read_input_tokens,
            }

        opik_usage = llm_usage.OpikUsage.from_anthropic_dict(usage_dict)
        return opik_usage
    except Exception:
        LOGGER.warning(
            "Failed to extract token usage from presumably Anthropic LLM vertexai langchain run.",
            exc_info=True,
        )
        return None


def is_anthropic_vertexai_run(run: "Run") -> bool:
    try:
        if run.serialized is None:
            return False

        invocation_params = run.extra.get("invocation_params", {})
        provider = invocation_params.get("_type", "").lower()
        is_anthropic_vertexai = (
            "vertexai" in provider.lower() and "anthropic" in provider.lower()
        )

        return is_anthropic_vertexai

    except Exception:
        LOGGER.debug(
            "Failed to check if Run instance is from Anthropic LLM vertexai, returning False.",
            exc_info=True,
        )
        return False


def _get_provider_and_model(
    run_dict: Dict[str, Any],
) -> Tuple[Optional[Union[Literal[LLMProvider.ANTHROPIC], str]], Optional[str]]:
    """
    Fetches the provider and model information from a given run dictionary.
    """
    provider = None
    model = None

    if invocation_params := run_dict["extra"].get("invocation_params"):
        provider = invocation_params.get("_type")
        if "vertexai" in provider.lower() and "anthropic" in provider.lower():
            provider = LLMProvider.ANTHROPIC_VERTEXAI
        model = invocation_params.get("model_name")

    return provider, model



from typing import Any, List, Optional, Union
from opik.evaluation.metrics.llm_judges.moderation import parser
import pydantic
from opik.evaluation.metrics import base_metric, score_result
from opik.evaluation.models import base_model, models_factory
from . import template


class ModerationResponseFormat(pydantic.BaseModel):
    score: float
    reason: str


class Moderation(base_metric.BaseMetric):
    """
    A metric that evaluates the moderation level of an input-output pair using an LLM.

    This metric uses a language model to assess the moderation level of the given input and output.
    It returns a score between 0.0 and 1.0, where higher values indicate more appropriate content.

    Args:
        model: The language model to use for moderation. Can be a string (model name) or an `opik.evaluation.models.OpikBaseModel` subclass instance.
            `opik.evaluation.models.LiteLLMChatModel` is used by default.
        name: The name of the metric. Defaults to "moderation_metric".
        few_shot_examples: A list of few-shot examples to be used in the query. If None, default examples will be used.
        track: Whether to track the metric. Defaults to True.
        project_name: Optional project name to track the metric in for the cases when
            there are no parent span/trace to inherit project name from.

    Example:
        >>> from opik.evaluation.metrics import Moderation
        >>> moderation_metric = Moderation()
        >>> result = moderation_metric.score("Hello, how can I help you?")
        >>> print(result.value)  # A float between 0.0 and 1.0
        >>> print(result.reason)  # Explanation for the score
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        name: str = "moderation_metric",
        few_shot_examples: Optional[List[template.FewShotExampleModeration]] = None,
        track: bool = True,
        project_name: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            track=track,
            project_name=project_name,
        )

        self._init_model(model)
        self.few_shot_examples = [] if few_shot_examples is None else few_shot_examples

    def _init_model(
        self, model: Optional[Union[str, base_model.OpikBaseModel]]
    ) -> None:
        if isinstance(model, base_model.OpikBaseModel):
            self._model = model
        else:
            self._model = models_factory.get(model_name=model)

    def score(self, output: str, **ignored_kwargs: Any) -> score_result.ScoreResult:
        """
        Calculate the moderation score for the given input-output pair.

        Args:
            output: The output text to be evaluated.
            **ignored_kwargs (Any): Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object containing the moderation score
            (between 0.0 and 1.0) and a reason for the score.
        """
        llm_query = template.generate_query(
            output=output, few_shot_examples=self.few_shot_examples
        )
        model_output = self._model.generate_string(
            input=llm_query, response_format=ModerationResponseFormat
        )

        return parser.parse_model_output(content=model_output, name=self.name)

    async def ascore(
        self, output: str, **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Asynchronously calculate the moderation score for the given input-output pair.

        This method is the asynchronous version of :meth:`score`. For detailed documentation,
        please refer to the :meth:`score` method.

        Args:
            output: The output text to be evaluated.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with the moderation score and reason.
        """

        llm_query = template.generate_query(
            output=output, few_shot_examples=self.few_shot_examples
        )
        model_output = await self._model.agenerate_string(
            input=llm_query, response_format=ModerationResponseFormat
        )

        return parser.parse_model_output(content=model_output, name=self.name)
